package SNMP_lib;

use strict;
use warnings;
use SNMP;
use Socket;
use Xorux_lib qw(error isdigit file_time_diff save_as_json read_json unobscure_password);
use Data::Dumper;

my $basedir  = $ENV{INPUTDIR} || Xorux_lib::error("INPUTDIR is not defined!") && exit;
my $tmpdir   = "$basedir/tmp";
my $mibs_dir = "$basedir/MIBs";
our $lan_cfg = "$basedir/etc/lan-list.cfg";
our $san_cfg = "$basedir/etc/san-list.cfg";

# SNMP parameters
our $community    = 'public';
our $snmp_version = '2c';
our $timeout      = '50000000';
our $remote_port  = "161";

#SNMP v3 parameters
our $sec_name;
our $sec_level;
our $sec_engine_id;
our $context_engine_id;
our $context;
our $auth_proto;
our $priv_proto;
our $auth_pass;
our $priv_pass;
my $auth_master_key;
my $priv_master_key;
my $auth_localized_key;
my $priv_localized_key;

sub initialize_snmp {
  my $cfg_file = shift;
  my $type     = shift;
  $type = $type ? $type : '';

  # SNMP version
  if ( defined $ENV{SNMP_VERSION} && $ENV{SNMP_VERSION} ne '' ) {
    if ( $type eq 'vsan-brcd' || $ENV{SNMP_VERSION} eq '3' ) {
      $snmp_version = '3';
      snmp_v3( $cfg_file, $type );
    }
    elsif ( $ENV{SNMP_VERSION} eq '2c' || $ENV{SNMP_VERSION} eq '1' ) {
      $snmp_version = $ENV{SNMP_VERSION};
    }
    else {
      Xorux_lib::error( "Unknown | $ENV{SNMP_VERSION} | SNMP version in $cfg_file! Automatically used SNMP version \"$snmp_version\"! $!" . __FILE__ . ":" . __LINE__ );
    }
  } ## end if ( defined $ENV{SNMP_VERSION...})

  # port
  if ( defined $ENV{SNMP_PORT} && $ENV{SNMP_PORT} ne '' ) {
    my $remote_port_cfg = $ENV{SNMP_PORT};

    if ( !Xorux_lib::isdigit( $remote_port_cfg ) ) {
      Xorux_lib::error( "SNMP port \"$remote_port_cfg\" is not digit! Automatically used SNMP default port \"161\"! $!" . __FILE__ . ":" . __LINE__ );
      if (  $remote_port_cfg < 65536 ) {
        $remote_port = $remote_port_cfg;
      }
      else {
        Xorux_lib::error( "SNMP port \"$remote_port_cfg\" is greater then 65536! Automatically used SNMP default port \"161\"! $!" . __FILE__ . ":" . __LINE__ );
      }
    }
  } ## end if ( defined $ENV{SNMP_PORT...})

  return 1;
} ## end sub initialize_snmp

sub initialize_mibs {
  my $type = shift;


  &SNMP::addMibDirs($mibs_dir);

  if ( $type eq 'LAN' ) {
    &SNMP::loadModules( 'SW-MIB', 'Brocade-REG-MIB', 'Brocade-TC', 'FCMGMT-MIB', 'CISCO-FC-FE-MIB',
                        'CISCO-DM-MIB', 'Q-BRIDGE-MIB', 'CISCO-VTP-MIB', 'CISCO-CDP-MIB', 'IEEE8023-LAG-MIB',
                        'CISCO-ETHERNET-FABRIC-EXTENDER-MIB', 'CISCO-STACKWISE-MIB', 'CISCO-VIRTUAL-SWITCH-MIB',
                        'CISCO-PAGP-MIB', 'CISCO-ENTITY-FRU-CONTROL-MIB', 'CISCO-VLAN-IFTABLE-RELATIONSHIP-MIB',
                        'CISCO-MEMORY-POOL-MIB', 'CISCO-PROCESS-MIB', 'CISCO-SYSTEM-EXT-MIB',
                        'CISCOSB-MIB', 'CISCOSB-rndMng' );
  }
  elsif ( $type eq 'LAN-discovery' ) {
    &SNMP::loadModules( 'CISCO-CDP-MIB' );
  }
  elsif ( $type eq 'SAN' ) {
    &SNMP::loadModules( 'SW-MIB', 'Brocade-REG-MIB', 'Brocade-TC', 'FCMGMT-MIB', 'FA-EXT-MIB', 'FIBRE-CHANNEL-FE-MIB', 'CISCO-FC-FE-MIB', 'CISCO-DM-MIB' );
  }
  elsif ( $type eq "VSPG"){
    &SNMP::loadModules( 'ALL' );
  }

  &SNMP::initMib();
  return 1;
} ## end sub initialize_mibs

sub initialize_community {
  my $ip       = shift;
  my $cfg_file = shift;
  my $type = shift;

  if ( -f $cfg_file && !defined $type ) {
    open( SL, '<', $cfg_file ) || Xorux_lib::error( "Couldn't open file $cfg_file $!" . __FILE__ . ":" . __LINE__ );
    my @lan_cfg = <SL>;
    close(SL);
    my ($line) = grep {/^$ip:/} @lan_cfg;
    chomp $line;
    my ( undef, $comm_st, undef, undef ) = split( ":", $line );
    if ( defined $comm_st && $comm_st ne "" ) {
      chomp $comm_st;
      $community = $comm_st;

      if ( defined $ENV{SNMP_VERSION} && $ENV{SNMP_VERSION} eq '3' && $cfg_file eq $san_cfg ) {
        $sec_name = $comm_st;
      }
    }
  }
  if (defined $type && $type eq "VSPG"){
    if (defined $ENV{SNMP_COMMUNITY} && $ENV{SNMP_COMMUNITY} ne ""){
      $community = $ENV{SNMP_COMMUNITY};
    }
    if (defined $ENV{SNMP_VERSION} && $ENV{SNMP_VERSION} eq '3' && defined $ENV{SNMP3_USER} ){
      $sec_name = $ENV{SNMP3_USER};
    }
  }
} ## end sub initialize_community

sub snmp_v3 {
  my $cfg_file = shift;
  my $type = shift;
  my $cfg;

  {
    local $/=undef;
    open ( FILE, '<', $cfg_file ) or return 0;
    $cfg = <FILE>;
    close FILE;
  }

  if (defined $type && $type eq "VSPG" ){
    if (!defined $sec_name){
      Xorux_lib::error( "SNMP v3 SecName must be defined! " . __FILE__ . ":" . __LINE__ );
    }
    if (defined $ENV{SNMP_SEC_LEVEL} && $ENV{SNMP_SEC_LEVEL} =~ m/^noAuthNoPriv$|^authNoPriv$|^authPriv$/){
      $sec_level = $ENV{SNMP_SEC_LEVEL};
    }
    else{
      Xorux_lib::error( "SNMP v3 SecLevel \"$ENV{SNMP_SEC_LEVEL}\" should be one of: 'noAuthNoPriv', 'authNoPriv' or 'authPriv'! " . __FILE__ . ":" . __LINE__ );
    }

    if (defined $ENV{SNMP_AUTH_PROTOCOL} && $ENV{SNMP_AUTH_PROTOCOL} =~ m/^SHA$|^MD5$/){
      $auth_proto = $ENV{SNMP_AUTH_PROTOCOL};
    }
    else{
      Xorux_lib::error( "SNMP v3 AuthProto \"$ENV{SNMP_AUTH_PROTOCOL}\" should be one of: 'MD5' or 'SHA'! " . __FILE__ . ":" . __LINE__ );
    }


    if (defined $ENV{SNMP_PRIV_PROTOCOL} && $ENV{SNMP_PRIV_PROTOCOL} =~ m/^DES$|^AES$/){
      $priv_proto = $ENV{SNMP_PRIV_PROTOCOL};
    }
    else{
      Xorux_lib::error( "SNMP v3 PrivProto \"$ENV{SNMP_PRIV_PROTOCOL}\" should be one of: 'DES' or 'AES'! " . __FILE__ . ":" . __LINE__ );
    }
    if ( defined $ENV{SNMP_AUTH_PASS} )         { $auth_pass = Xorux_lib::unobscure_password($ENV{SNMP_AUTH_PASS}); }
    if ( defined $ENV{SNMP_PRIV_PASS} )         { $priv_pass = Xorux_lib::unobscure_password($ENV{SNMP_PRIV_PASS}); }

  }
  else{

    if ( defined $sec_name ) {
      ## already defined as part of switch configuration in place of community (snmp v3), hack
    }
    elsif ( $cfg =~ /^SEC_NAME=(.+)$/m ) {
      $sec_name = $1;
    }
    elsif ( defined $ENV{SAN_TYPE} && ( $ENV{SAN_TYPE} eq 'BRCD' || $ENV{SAN_TYPE} eq 'QLOGIC' ) ) {
      $sec_name = 'snmpuser1';
    }
    else {
      Xorux_lib::error( "SNMP v3 SecName must be defined! " . __FILE__ . ":" . __LINE__ );
    }

    if ( $cfg =~ /^SEC_LEVEL=(.+)$/m ) {
      if ( $1 eq 'noAuthNoPriv' || $1 eq 'authNoPriv' || $1 eq 'authPriv' ) {
        $sec_level = $1;
      }
      else {
        Xorux_lib::error( "SNMP v3 SecLevel \"$1\" should be one of: 'noAuthNoPriv', 'authNoPriv' or 'authPriv'! " . __FILE__ . ":" . __LINE__ );
      }
    }
    if ( $cfg =~ /^AUTH_PROTOCOL=(.+)$/m ) {
      if ( $1 eq 'MD5' || $1 eq 'SHA' ) {
        $auth_proto = $1;
      }
      else {
        Xorux_lib::error( "SNMP v3 AuthProto \"$1\" should be one of: 'MD5' or 'SHA'! " . __FILE__ . ":" . __LINE__ );
      }
    }
    if ( $cfg =~ /^PRIV_PROTOCOL=(.+)$/m ) {
      if ( $1 eq 'DES' || $1 eq 'AES' ) {
        $priv_proto = $1;
      }
      else {
        Xorux_lib::error( "SNMP v3 AuthProto \"$1\" should be one of: 'DES' or 'AES'! " . __FILE__ . ":" . __LINE__ );
      }
    }
    if ( $cfg =~ /^SEC_ENGINE_ID=(.+)$/m ) {
      my $id = $1;

      if ( $id =~ /[0-9a-fA-F]{10,64}/ ) {
        $sec_engine_id = $1;
      }
      else {
        Xorux_lib::error( "SNMP v3 SecEngineId \"$id\" should hexadecimal number with length in range from 10 to 64! " . __FILE__ . ":" . __LINE__ );
      }
    }
    if ( $cfg =~ /^AUTH_PASS=(.+)$/m )         { $auth_pass = Xorux_lib::unobscure_password($1); }
    if ( $cfg =~ /^PRIV_PASS=(.+)$/m )         { $priv_pass = Xorux_lib::unobscure_password($1); }
    if ( $cfg =~ /^CONTEXT_ENGINE_ID=(.+)$/m )  { $context_engine_id = $1; }
    if ( $cfg =~ /^CONTEXT=(.+)$/m )          { $context = $1; }
    if ( $cfg =~ /^AUTH_MASTER_KEY=(.+)$/m )    { $auth_master_key = $1; }
    if ( $cfg =~ /^PRIV_MASTER_KEY=(.+)$/m )    { $priv_master_key = $1; }
    if ( $cfg =~ /^AUTH_LOCALIZED_KEY=(.+)$/m ) { $auth_localized_key = $1; }
    if ( $cfg =~ /^PRIV_LOCALIZED_KEY=(.+)$/m ) { $priv_localized_key = $1; }

  }

  return 1;
}

sub new_session {
  my $params          = shift;
  my $community_index = shift;
  my %snmpparms       = defined $params ? %{ $params } : ();
  my $sess;

  $snmpparms{DestHost}       = inet_ntoa( inet_aton( $snmpparms{DestHost} ) );
  $snmpparms{Version}        = defined $snmpparms{Version} ? $snmpparms{Version} : $snmp_version;
  $snmpparms{Timeout}        = defined $snmpparms{Timeout} ? $snmpparms{Timeout} : $timeout;
  $snmpparms{RemotePort}     = defined $snmpparms{RemotePort} ? $snmpparms{RemotePort} : $remote_port;
  $snmpparms{UseSprintValue} = '1';

  if ( $snmp_version eq '2c' ) {
    my $community_string = $community_index ? "$community\@$community_index" : $community;
    $snmpparms{Community} = defined $snmpparms{Community} ? $snmpparms{Community} : $community_string;
  }
  elsif ( $snmp_version eq '3' ) {
    $snmpparms{SecName} = $sec_name;
    my $session_context = $context;

    if ( defined $ENV{LAN_TYPE} && $ENV{LAN_TYPE} eq 'CISCO' && $community_index ) {
      $session_context = "vlan-$community_index";
    }

    if ( defined $sec_level )          { $snmpparms{SecLevel}         = $sec_level; }
    if ( defined $sec_engine_id )      { $snmpparms{SecEngineId}      = $sec_engine_id; }
    if ( defined $context_engine_id )  { $snmpparms{ContextEngineId}  = $context_engine_id; }
    if ( defined $session_context )    { $snmpparms{Context}          = $session_context; }
    if ( defined $auth_pass )          { $snmpparms{AuthPass}         = $auth_pass; }
    if ( defined $priv_pass )          { $snmpparms{PrivPass}         = $priv_pass; }
    if ( defined $auth_proto )         { $snmpparms{AuthProto}        = $auth_proto; }
    if ( defined $priv_proto )         { $snmpparms{PrivProto}        = $priv_proto; }
    if ( defined $auth_master_key )    { $snmpparms{AuthMasterKey}    = $auth_master_key; }
    if ( defined $priv_master_key )    { $snmpparms{PrivMasterKey}    = $priv_master_key; }
    if ( defined $auth_localized_key ) { $snmpparms{AuthLocalizedKey} = $auth_localized_key; }
    if ( defined $priv_localized_key ) { $snmpparms{PrivLocalizedKey} = $priv_localized_key; }
  }

  $sess = new SNMP::Session(%snmpparms);

  return $sess;
} ## end sub snmp_session

sub check_mib_compatibility {
  my $dest               = shift;
  my $vendor             = shift;
  my $compatibility_file = "$tmpdir/$dest-mib-compatiblity.txt";
  my %compatibility      = ();

  ## check mib compatibility only once per 1 hour
  if ( !-f $compatibility_file || Xorux_lib::file_time_diff( $compatibility_file ) > 3600 ) {
    my $sess = new_session( { DestHost => $dest, UseLongNames => 1 } );

    $compatibility{"Q-BRIDGE-MIB"} = test_mib( $sess, 'qBridgeMIB' );
    $compatibility{"LLDP-MIB"}     = test_mib( $sess, 'lldpMIB' );

    if ( $vendor eq 'CISCO' ) {
      $compatibility{"CISCO-CDP-MIB"}                       = test_mib( $sess, 'ciscoCdpMIB' );
      $compatibility{"CISCO-VTP-MIB"}                       = test_mib( $sess, 'ciscoVtpMIB' );
      $compatibility{"CISCO-PAGP-MIB"}                      = test_mib( $sess, 'ciscoPagpMIB' );
      $compatibility{"CISCO-VIRTUAL-SWITCH-MIB"}            = test_mib( $sess, 'ciscoVirtualSwitchMIB' );
      $compatibility{"CISCO-STACKWISE-MIB"}                 = test_mib( $sess, 'ciscoStackWiseMIB' );
      $compatibility{"CISCO-ETHERNET-FABRIC-EXTENDER-MIB"}  = test_mib( $sess, 'ciscoEthernetFabricExtenderMIB' );
      $compatibility{"CISCO-PROCESS-MIB"}                   = test_mib( $sess, 'ciscoProcessMIB' );
      $compatibility{"CISCO-MEMORY-POOL-MIB"}               = test_mib( $sess, 'ciscoMemoryPoolMIB' );
      $compatibility{"CISCO-SYSTEM-EXT-MIB"}                = test_mib( $sess, 'ciscoSystemExtMIB' );
      $compatibility{"CISCO-ENTITY-FRU-CONTROL-MIB"}        = test_mib( $sess, 'ciscoEntityFRUControlMIB' );
      $compatibility{"CISCO-VLAN-IFTABLE-RELATIONSHIP-MIB"} = test_mib( $sess, 'ciscoVlanIfTableRelationshipMIB' );
      $compatibility{"CISCOSB-rndMng"}                      = test_mib( $sess, 'rndMng' );

      # test connection with community string index on vlan 1
      $sess = new_session( { DestHost => $dest, UseLongNames => 1 }, 1 );
      $compatibility{"COMMUNITY-STRING-INDEX"}    = test_mib( $sess, 'dot1dBridge' );
    }

    Xorux_lib::save_as_json( $compatibility_file, \%compatibility );
  }
  else {
    %compatibility = %{ Xorux_lib::read_json( $compatibility_file ) };
  }

  return \%compatibility;
} ## end sub check_mib_compatibility

sub collect {
  my $data               = shift;
  my $sess               = shift;
  my $section            = shift;
  my $mib                = shift;
  my $transformation_sub = shift;
  my $iid_index          = shift; ## used for wrongly indexed entries with leading or trailing indexes
  my $vb                 = new SNMP::Varbind( [$mib] );
  my %output;

  if ( defined $section && $section eq 'configuration') {
    my $var = $sess->getnext($vb);

    if ( defined $var ) {
      $var =~ s/"//g;
      unless ( $var eq '' ) {
        $data->{configuration}{$mib} = $var;
      }
    }
  }
  else {
    for ( my $var = $sess->getnext($vb); ( $vb->tag eq $mib ) and not( $sess->{ErrorNum} ); $var = $sess->getnext($vb) ) {
      my $index = defined $iid_index ? (split /\./, $vb->iid)[$iid_index] : $vb->iid;
      $var =~ s/"//g;
      $var = defined $transformation_sub ? $transformation_sub->($var) : $var;

      if ( $var eq '' ) { next; }

      # get data into hash and return it as reference
      $output{ $index } = $var;

      # get indexed data into data hash
      if ( defined $section ) {
        $data->{ $section }{ $index }{ $mib } = $var;
      }
    } ## end for ( $var = $sess->getnext...)

  }
  if ( $sess->{ErrorNum} ) {
    Xorux_lib::error( "Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib. $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }

  return \%output;
} ## end sub snmp_collect

sub collect_san {
  my $data               = shift;
  my $sess               = shift;
  my $section            = shift;
  my $mib                = shift;
  my $mib_alias          = shift;
  my $vb                 = new SNMP::Varbind( [$mib] );
  my %output;

  if ( ! defined $mib_alias || $mib_alias eq '' ) { $mib_alias = $mib; }

  if ( defined $ENV{DEBUG_SAN} && $ENV{DEBUG_SAN} eq "1" ) { print localtime() . " : $mib : $mib_alias : START\n"; }

  if ( defined $section && $section eq 'configuration') {
    my $var = $sess->getnext($vb);

    if ( defined $var ) {
      $var =~ s/"//g;
      $var =~ s/^\s+//g;
      $var =~ s/\s+$//g;
      unless ( $var eq '' ) {
        $data->{configuration}{$mib_alias} = $var;
      }
    }
  }
  else {
    for ( my $var = $sess->getnext($vb); ( $vb->tag eq $mib ) and not( $sess->{ErrorNum} ); $var = $sess->getnext($vb) ) {
      my $index = $vb->iid;
      if ( $mib ne "zoneEnforcedZoneName" && $mib ne "zoneEnforcedZoneMemberID" && $mib ne "zoneSetName" && $mib ne "fcIfNxPortName" && $mib ne "fcFxPortNxPortName" ) {
        $index =~ s/^.+\.//g;
      }
      $var =~ s/"//g;
      $var =~ s/^\s+//g;
      $var =~ s/\s+$//g;

      if ( $var eq '' ) { next; }

      # get data into hash and return it as reference
      $output{ $index } = $var;

      # get indexed data into data hash
      if ( defined $section ) {
        $data->{ $section }{ $index }{ $mib_alias } = $var;
      }
    } ## end for ( $var = $sess->getnext...)

  }
  if ( $sess->{ErrorNum} ) {
    Xorux_lib::error( "Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib. $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }

  if ( defined $ENV{DEBUG_SAN} && $ENV{DEBUG_SAN} eq "1" ) { print localtime() . " : $mib : $mib_alias : END\n"; }

  return \%output;
}

sub collect_selected {
  my $data               = shift;
  my $sess               = shift;
  my @selected_indexes   = @{ shift @_ };
  my $section            = shift;
  my $mib                = shift;
  my $transformation_sub = shift;
  my $iid_index          = shift; ## used for wrongly indexed entries with leading or trailing indexes
  my %output;
  my $var;
  my $vb;

  foreach my $selected_index ( @selected_indexes ) {
    $vb = new SNMP::Varbind( [ $mib, $selected_index ] );
    $var = $sess->get($vb);

    if ( ( $vb->tag eq $mib ) and not ( $sess->{ErrorNum} ) and
         ( $var ne 'No Such Instance currently exists at this OID' ) ) {
      my $index = defined $iid_index ? (split /\./, $vb->iid)[$iid_index] : $vb->iid;
      $var =~ s/"//g;
      $var = defined $transformation_sub ? $transformation_sub->($var) : $var;

      if ( $var eq '' ) { next; }

      # get data into hash and return it as reference
      $output{ $index } = $var;

      # get indexed data into data hash
      if ( defined $section ) {
        $data->{ $section }{ $index }{ $mib } = $var;
      }
    }
  }

  if ( $sess->{ErrorNum} ) {
    Xorux_lib::error( "Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib. $!" . __FILE__ . ":" . __LINE__ ) && exit;
    return 0;
  }

  return \%output;
} ## end sub snmp_collect_selected

sub snmp_fetch {
  my $sess = shift;
  my $mib  = shift;
  my $vb   = new SNMP::Varbind( [$mib] );
  my $var  = $sess->getnext($vb);

  if ( ( $vb->tag =~ /$mib/ ) and not( $sess->{ErrorNum} ) ) {
    $var =~ s/"//g;
    return $var;
  }

  if ( $sess->{ErrorNum} ) {
    Xorux_lib::error( "Got $sess->{ErrorStr} querying $sess->{DestHost} with $sess->{Community} for $mib. $!" . __FILE__ . ":" . __LINE__ ) && return undef;
  }

  return undef;
} ## end sub snmp_fetch

sub test_mib {
  my $sess   = shift;
  my $mib    = shift;
  my $vb     = new SNMP::Varbind( [$mib] );
  my $result = $sess->getnext($vb);

  if ( $vb->tag !~ /\.$mib\./ || $sess->{ErrorNum} ) {
    return 0;
  }

  return 1;
} ## end sub test_mib

sub configuration_string {
  my $str = '';
  my $snmp_version_str = $snmp_version eq '2c' ? '2c (default)' : $snmp_version;
  my $remote_port_str  = $remote_port eq '161' ? '161 (default)' : $remote_port;
  my $timeout_str      = $timeout / 1000000 . " seconds";

  $str .= "SNMP version  : $snmp_version_str\n";
  $str .= "Port          : $remote_port_str\n";
  $str .= "Timeout       : $timeout_str\n";

  if ( $SNMP_lib::snmp_version eq '3' ) {
    my $sec_name_str   = defined $sec_name ? $sec_name : '';
    my $sec_level_str  = defined $sec_level ? $sec_level : 'noAuthNoPriv (default)';
    my $auth_proto_str = defined $auth_proto ? $auth_proto : 'MD5 (default)';
    my $priv_proto_str = defined $priv_proto ? $priv_proto : 'DES (default)';

    $str .= "SecName       : $sec_name_str\n";
    $str .= "SecLevel      : $sec_level_str\n";
    $str .= "AuthProto     : $auth_proto_str\n";
    $str .= "PrivProto     : $priv_proto_str\n";
  }
  else {
    my $community_str = defined $community ? $community : '';
    $str .= "Community     : $community_str\n";
  }

  return \$str;
}

1;
