#!/usr/bin/perl

use strict;
use warnings;
use SNMP;
use Socket;
use Data::Dumper;
use Time::Local;
use Math::BigInt;
use SNMP_lib qw(initialize_snmp initialize_community check_mib_compatibility collect collect_selected new_session);
use Xorux_lib qw(error isdigit file_time_diff save_as_json read_json unobscure_password);

my $lan_ip       = $ENV{LAN_IP}   || Xorux_lib::error("LAN IP is required!")      && exit;
my $lan_type     = $ENV{LAN_TYPE} ? $ENV{LAN_TYPE} : "UNKNOWN";
my $lan_group    = $ENV{LAN_GROUP};
my $basedir      = $ENV{INPUTDIR} || Xorux_lib::error("INPUTDIR is not defined!") && exit;
my $webdir       = $ENV{WEBDIR}   ||= "$basedir/www" ;
my $wrkdir       = "$basedir/data";
my $tmpdir       = "$basedir/tmp";
my $lan_list_cfg = "$basedir/etc/lan-list.cfg";
my $cisco        = $lan_type eq 'CISCO';
my $update_conf  = 0;
my $switch_name;
my $switch_dir;
my %mib_compatibility;
my %data;
my %conf;

# LAN data collection timeout
my $lan_data_collection_timeout = 600;
if ( defined $ENV{LAN_DATA_COLLECTION_TIMEOUT} && Xorux_lib::isdigit( $ENV{LAN_DATA_COLLECTION_TIMEOUT} ) ) {
  $lan_data_collection_timeout = $ENV{LAN_DATA_COLLECTION_TIMEOUT};
}

eval {
  # Set alarm
  my $act_time = localtime();
  local $SIG{ALRM} = sub { die "$act_time: died in SIG ALRM"; };
  alarm($lan_data_collection_timeout);
  SNMP_lib::initialize_snmp( $lan_list_cfg );
  SNMP_lib::initialize_mibs( 'LAN' );
  SNMP_lib::initialize_community( $lan_ip, $lan_list_cfg );
  %mib_compatibility = %{ SNMP_lib::check_mib_compatibility( $lan_ip, $lan_type ) };

  my $sess = SNMP_lib::new_session( { DestHost => $lan_ip } );

  collect_data($sess);
  collect_configuration($sess);
  create_switch_directories();
  save_performance_data();
  save_configuration();
  save_health_status();

  # end of alarm
  alarm(0);
};
if ($@) {
  if ( $@ =~ /died in SIG ALRM/ ) {
    Xorux_lib::error( "Data collection via snmp timed out after $lan_data_collection_timeout seconds : " . __FILE__ . ":" . __LINE__ );
    exit(0);
  }
}

exit;


sub collect_data {
  my $sess = shift; # The SNMP::Session object that does the work.

  SNMP_lib::collect(\%data, $sess, 'configuration', 'sysName');

  if ( defined $data{configuration}{sysName} ) {
    $switch_name = $data{configuration}{sysName};
  }
  else {
    Xorux_lib::error( "Switch name not found! IP: $lan_ip $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }

  SNMP_lib::collect(\%data, $sess, 'ports', 'ifName');
  SNMP_lib::collect(\%data, $sess, 'ports', 'ifAlias');
  SNMP_lib::collect(\%data, $sess, 'ports', 'ifType');
  SNMP_lib::collect(\%data, $sess, 'ports', 'ifOperStatus');
  SNMP_lib::collect(\%data, $sess, 'ports', 'ifAdminStatus');
  SNMP_lib::collect(\%data, $sess, 'ports', 'ifHighSpeed');

  SNMP_lib::collect(\%data, $sess, 'ports', 'ifHCOutOctets', \&bignum); # bytes transmitted - data out
  SNMP_lib::collect(\%data, $sess, 'ports', 'ifHCInOctets', \&bignum); # bytes recived - data in
  SNMP_lib::collect(\%data, $sess, 'ports', 'ifOutUcastPkts'); # Unicast Packets transmitted - Ucast_pkts_out
  SNMP_lib::collect(\%data, $sess, 'ports', 'ifInUcastPkts'); # Unicast Packets recived - Ucast_pkts_in
  SNMP_lib::collect(\%data, $sess, 'ports', 'ifOutErrors'); # transmitted Packets with Errors - errors_out
  SNMP_lib::collect(\%data, $sess, 'ports', 'ifInErrors'); # recived Packets with Errors - errors_i
#  SNMP_lib::collect(\%data, $sess, 'ports', 'ifOutMulticastPkts'); # Multicast Packets trans. - Mcast_pkts_out
#  SNMP_lib::collect(\%data, $sess, 'ports', 'ifInMulticastPkts'); # Multicast Packets recived - Mcast_pkts_in
#  SNMP_lib::collect(\%data, $sess, 'ports', 'ifOutBroadcastPkts'); # Broadcast Packets trans. - Bcast_pkts_out
#  SNMP_lib::collect(\%data, $sess, 'ports', 'ifInBroadcastPkts'); # Broadcast Packets recived - Bcast_pkts_in

  if ( $cisco ) {
    if ( $mib_compatibility{'CISCO-ENTITY-FRU-CONTROL-MIB'} ) {
      SNMP_lib::collect(\%data, $sess, 'switch_status', 'cefcModuleAdminStatus');
      my @fru_indexes = keys %{ SNMP_lib::collect(\%data, $sess, 'switch_status', 'cefcModuleOperStatus') };

      SNMP_lib::collect_selected(\%data, $sess, \@fru_indexes, 'physical_entries', 'entPhysicalDescr');
      SNMP_lib::collect_selected(\%data, $sess, \@fru_indexes, 'physical_entries', 'entPhysicalSerialNum');
    }
    ## fex ports
    if ( $mib_compatibility{'CISCO-ETHERNET-FABRIC-EXTENDER-MIB'} ) {
      SNMP_lib::collect(\%data, $sess, 'fex_ports', 'cefexBindingExtenderIndex');
    }

    if ( $mib_compatibility{'CISCO-PROCESS-MIB'} ) {
      SNMP_lib::collect(\%data, $sess, 'cpu', 'cpmCPUTotal5minRev');
      my @cpu_indexes = values %{ SNMP_lib::collect(\%data, $sess, 'cpu', 'cpmCPUTotalPhysicalIndex') };

      SNMP_lib::collect_selected(\%data, $sess, \@cpu_indexes, 'physical_entries', 'entPhysicalName');
    }

    if ( $mib_compatibility{'CISCO-MEMORY-POOL-MIB'} ) {
      SNMP_lib::collect(\%data, $sess, 'memory', 'ciscoMemoryPoolName');
      SNMP_lib::collect(\%data, $sess, 'memory', 'ciscoMemoryPoolUsed');
      SNMP_lib::collect(\%data, $sess, 'memory', 'ciscoMemoryPoolFree');
    }

    if ( $mib_compatibility{"CISCO-SYSTEM-EXT-MIB"} ) {
      SNMP_lib::collect(\%data, $sess, 'memory', 'cseSysMemoryUtilization');
    }

    if ( $mib_compatibility{"CISCOSB-rndMng"} ) {
      SNMP_lib::collect(\%data, $sess, 'cpu', 'rlCpuUtilDuringLast5Minutes');
      my @cpu_indexes = values %{ SNMP_lib::collect(\%data, $sess, 'cpu', 'cpmCPUTotalPhysicalIndex') };

      SNMP_lib::collect_selected(\%data, $sess, \@cpu_indexes, 'physical_entries', 'entPhysicalName');
    }
  }

  return 1;
} ## end sub collect_data

sub create_switch_directories {
  my $act_time = time();
  $switch_dir  = "$wrkdir/$switch_name";

  if ( !-d $switch_dir ) {
    mkdir( $switch_dir, 0755 ) || Xorux_lib::error( "$act_time: Cannot mkdir $switch_dir: $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }
  if ( !-d "$webdir/$switch_name" ) {
    mkdir( "$webdir/$switch_name", 0755 ) || Xorux_lib::error( "$act_time: Cannot mkdir $webdir/$switch_name: $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }

  my $marker_file = "$switch_dir/LAN-$lan_type";
  if ( !-f $marker_file ) {
    open( CISC, '>', $marker_file ) || Xorux_lib::error( "Couldn't open file $marker_file $!" . __FILE__ . ":" . __LINE__ ) && exit;
    close(CISC);
  }

  my $group_file = "$switch_dir/GROUP.cfg";
  if ( defined $lan_group && $lan_group ne '' ) {
    open( GROUP, '>', $group_file ) || Xorux_lib::error( "Couldn't open file $group_file $!" . __FILE__ . ":" . __LINE__ ) && exit;
    print GROUP $lan_group;
    $conf{switch}{switch_group} = $lan_group;
    close(GROUP);
  }
  elsif ( -f $group_file ) {
    unlink( $group_file );
  }

  return 1;
} ## sub create_switch_directories

sub save_performance_data {
  my ( $sec, $min, $hour, $day, $month, $year, $wday, $yday, $isdst ) = localtime();
  my $act_time    = time();
  my $date        = sprintf( "%4d%02d%02d", $year + 1900, $month + 1, $day );
  my $time        = sprintf( "%02d%02d", $hour, $min );
  my $output_file = "$switch_dir/$switch_name\_lanperf_$date\_$time.out.tmp";

  open( my $out, '>', $output_file ) || Xorux_lib::error( "Couldn't open file $output_file $!" . __FILE__ . ":" . __LINE__ ) && exit;

  print $out "#PORT,act_time,port_id,data_out,data_in,unicast_pkts_out,unicast_pkts_in,multicast_pkts_out,multicast_pkts_in,broadcast_pkts_out,broadcast_pkts_in,errors_out,errors_in,reserve1,reserve2,switch_name,db_name\n";

  #
  # PORTS.cfg
  #
  my $ports_cfg        = "$switch_dir/PORTS.cfg";
  my $ports_cfg_suffix = "$ports_cfg-tmp";
  open( my $pcfg, '>', $ports_cfg_suffix ) || Xorux_lib::error( "Couldn't open file $ports_cfg_suffix $!" . __FILE__ . ":" . __LINE__ ) && exit;

  #
  # port configuration
  #
  foreach my $index ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
    my %port = %{ $data{ports}{ $index } };
    my $db_name      = "";
    my $port_id      = "";
    my $port_alias   = "";
    my $port_type    = "";
    my $port_subtype = "";

    if ( defined $port{ifName} ) {
      $port_id = name_format( $port{ifName} );
      $db_name = $port_id . ".rrd";
    }
    else { next; }

    if ( defined $port{ifType} ) {
      $port_type = port_type( $port{ifName}, $port{ifType} );

      unless ( $port_type ) { next; }
    }
    else { next; }

    if ( $cisco && $data{fex_ports}{ $index } ) {
      $port_subtype = 'fexport';
      $conf{ports}{ $index }{port_subtype} = $port_subtype;
    }

    if ( $port_type eq 'vlan' ) {
      $port_subtype = 'routed';

      if ( $cisco && $port{ifName} =~ /^VLAN/ ) {
        $port_subtype = 'unrouted';
      }
      $conf{ports}{ $index }{port_subtype} = $port_subtype;
    }

    $conf{ports}{ $index }{port_id}   = $port_id;
    $conf{ports}{ $index }{port_type} = $port_type;

    if ( $port_type eq 'stackport' ) {
      next;
    }

    if ( defined $port{ifAlias} && $port{ifAlias} ne '' ) {
      $port_alias = $port{ifAlias};
    }

    #
    # port performance stats
    #
    my $data_out            = Xorux_lib::isdigit( $port{ifHCOutOctets} )  ? $port{ifHCOutOctets}  : '';
    my $data_in             = Xorux_lib::isdigit( $port{ifHCInOctets} )   ? $port{ifHCInOctets}   : '';
    my $unicast_pkts_out    = Xorux_lib::isdigit( $port{ifOutUcastPkts} ) ? $port{ifOutUcastPkts} : '';
    my $unicast_pkts_in     = Xorux_lib::isdigit( $port{ifInUcastPkts} )  ? $port{ifInUcastPkts}  : '';
    my $errors_out          = Xorux_lib::isdigit( $port{ifOutErrors} )    ? $port{ifOutErrors}    : '';
    my $errors_in           = Xorux_lib::isdigit( $port{ifInErrors} )     ? $port{ifInErrors}     : '';
    my $multicast_pkts_out  = '';
    my $multicast_pkts_in   = '';
    my $broadcast_pkts_out  = '';
    my $broadcast_pkts_in   = '';
    my $reserve1            = '';
    my $reserve2            = '';

    # filter non-used ports
    if ( $data_out ne '0' && $data_out ne '' && $data_in ne '0' && $data_in ne '' ) {
      print $pcfg "$port_id : $db_name : $port_alias : $port_type : $port_subtype\n";
    }

    print $out "PORT,$act_time,$port_id,$data_out,$data_in,$unicast_pkts_out,$unicast_pkts_in,$multicast_pkts_out,$multicast_pkts_in,$broadcast_pkts_out,$broadcast_pkts_in,$errors_out,$errors_in,$reserve1,$reserve2,$switch_name,$db_name\n";
  } ## end foreach my $name ( sort ... ports )

  if ( $cisco ) {
    cpu_cisco( $out );
    memory_cisco( $out );
  }

  close($out);
  close($pcfg);
  rename "$ports_cfg_suffix", "$ports_cfg";

  return 1;
} ## sub save_performance_data

sub collect_configuration {
  my $sess            = shift;
  my $conf_touch_file = "$tmpdir/$lan_ip-snmp-conf.touch";

  ## collect configuration only once per hour
  if ( !-f $conf_touch_file || Xorux_lib::file_time_diff($conf_touch_file) > 3600 ) {
    my $ret = `touch $conf_touch_file`;

    if ( $ret ) {
      Xorux_lib::error( "$ret " . __FILE__ . ":" . __LINE__ );
      return 0;
    }
    $update_conf = 1;
  }
  else {
    return 1;
  }

  ## switch configurations
  SNMP_lib::collect(\%data, $sess, 'configuration', 'sysLocation');
  SNMP_lib::collect(\%data, $sess, 'configuration', 'sysUpTime');
  SNMP_lib::collect(\%data, $sess, 'physical_entries', 'entPhysicalClass');
  SNMP_lib::collect(\%data, $sess, 'physical_entries', 'entPhysicalName');
  SNMP_lib::collect(\%data, $sess, 'physical_entries', 'entPhysicalSerialNum');
  SNMP_lib::collect(\%data, $sess, 'physical_entries', 'entPhysicalModelName');
  SNMP_lib::collect(\%data, $sess, 'physical_entries', 'entPhysicalSoftwareRev');

  ## port configurations
  SNMP_lib::collect(\%data, $sess, 'ports', 'ifLastChange');
  SNMP_lib::collect(\%data, $sess, 'port_channels', 'dot3adAggPortSelectedAggID');

  ## vlans
  if ( $mib_compatibility{"Q-BRIDGE-MIB"} ) {
    SNMP_lib::collect(\%data, $sess, 'vlans', 'dot1qVlanFdbId', undef, 1);
    SNMP_lib::collect(\%data, $sess, 'vlans', 'dot1qVlanCurrentEgressPorts', undef, 1);
    SNMP_lib::collect(\%data, $sess, 'vlans', 'dot1qVlanStaticName');
  }

  ## lldp
  if ( $mib_compatibility{"LLDP-MIB"} && !$mib_compatibility{"CISCO-CDP-MIB"} ) {
    SNMP_lib::collect(\%data, $sess, 'lldp', 'lldpRemChassisIdSubtype');
    SNMP_lib::collect(\%data, $sess, 'lldp', 'lldpRemChassisId');
    SNMP_lib::collect(\%data, $sess, 'lldp', 'lldpRemPortId');
    SNMP_lib::collect(\%data, $sess, 'lldp', 'lldpRemSysName');
    SNMP_lib::collect(\%data, $sess, 'lldp', 'lldpRemSysCapEnabled', \&parse_capabilities_lldp);
  }

  ## macs
  SNMP_lib::collect(\%data, $sess, 'mac_entries', 'dot1dTpFdbAddress', \&mac_format);
  SNMP_lib::collect(\%data, $sess, 'mac_entries', 'dot1dTpFdbPort');
  SNMP_lib::collect(\%data, $sess, 'mac_index', 'dot1dBasePortIfIndex');

  if ( $lan_type eq "CISCO" ) {
    collect_configuration_cisco( $sess );
  }

  return 1;
}

sub save_configuration {
  if ( $update_conf ) {
    if ( keys %{ $data{ports} } ) { ports_conf(); }
    if ( keys %{ $data{vlans} } || keys %{ $data{vlan_member_ports} } ) { vlans(); }
    if ( keys %{ $data{port_channels} } ) { port_channels(); }
    if ( keys %{ $data{lldp} } ) { topology(); }
    if ( $cisco ) {
      if ( keys %{ $data{fex_config} } ) { fex_cisco(); }
      if ( keys %{ $data{vss_chassis} } ) { vss_cisco(); }
      if ( keys %{ $data{cdp} } ) { topology(); }
      if ( keys %{ $data{stack} } ) { stack_cisco(); }
    }
    # config.cfg
    if ( scalar keys %{ $data{configuration} } ) { switch_conf(); }

    Xorux_lib::save_as_json( "$switch_dir/config.cfg", \%conf);
  }

  return 1;
} ## end sub save_configurations

sub ports_conf {
  my $up_time = $data{configuration}{sysUpTime};

  read_alias_cfg();

  ## ports conf
  foreach my $index ( keys %{ $data{ports} } ) {
    my %port           = %{ $data{ports}{ $index } };
    my $port_name      = $port{ifName};
    my $speed          = $port{ifHighSpeed};
    my $alias_sw       = $port{ifAlias};
    my $alias_cfg      = $port{cfgAlias};
    my $operStatus     = $port{ifOperStatus};
    my $last_change    = $port{ifLastChange};
    my $admin_status   = $port{ifAdminStatus};

    if ( defined $port_name && defined $conf{ports}{ $index }{port_type} ) {
      $conf{ports}{ $index }{port_name} = $port_name;
    }
    else {
      next;
    }

    if ( defined $speed )        { $conf{ports}{ $index }{speed} = $speed;}
    if ( defined $alias_sw )     { $conf{ports}{ $index }{alias_from_sw} = $alias_sw; }
    if ( defined $alias_cfg )    { $conf{ports}{ $index }{alias_from_cfg} = $alias_cfg; }
    if ( defined $alias_sw )     { $conf{ports}{ $index }{alias_from_sw} = $alias_sw; }
    if ( defined $admin_status ) { $conf{ports}{ $index }{admin_status} = $admin_status; }

    if ( defined $operStatus ) {
      $conf{ports}{ $index }{state} = $operStatus eq 'up' ? "used" : "unused"; }
    elsif ( defined $last_change && defined $up_time ) {
      $conf{ports}{ $index }{state} = happened_before( $up_time, $last_change ) ? "used" : "unused";
    }
  }

  ## mac addresses
  foreach my $key ( keys %{ $data{mac_entries} } ) {
    my $mac_address = $data{mac_entries}{ $key }{dot1dTpFdbAddress};
    my $bridge_port = $data{mac_entries}{ $key }{dot1dTpFdbPort};
    unless ( defined $mac_address && defined $bridge_port ) { next; };

    my $port_index  = $data{mac_index}{ $bridge_port }{dot1dBasePortIfIndex};
    unless ( defined $port_index ) { next; };

    my $port_name   = $data{ports}{$port_index}{ifName};
    unless ( defined $port_name ) { next; };

    push @{ $conf{ports}{ $port_index }{macs} }, $mac_address;
  }

  return 1;
} ## end sub ports_conf

sub topology {
  foreach my $index ( keys %{ $data{lldp} } ) {
    my %lldp_entry   = %{ $data{lldp}{$index} };
    my @capabilities = defined $lldp_entry{lldpRemSysCapEnabled} ? @{ $lldp_entry{lldpRemSysCapEnabled} } : ();
    my $if_index     = ( split /\./, $index )[1];

    unless ( defined $lldp_entry{lldpRemSysCapEnabled} && scalar grep(/^bridge$/, @{ $lldp_entry{lldpRemSysCapEnabled} } ) ) {
      next;
    }

    if ( defined $lldp_entry{lldpRemChassisIdSubtype} ) {
      $conf{topology}{ $if_index }{address_type} = $lldp_entry{lldpRemChassisIdSubtype};
    }
    if ( defined $lldp_entry{lldpRemChassisId} ) {
      $conf{topology}{ $if_index }{address} = $lldp_entry{lldpRemChassisId};
    }
    if ( defined $lldp_entry{lldpRemPortId} ) {
      $conf{topology}{ $if_index }{remote_port} = $lldp_entry{lldpRemPortId};
    }
    if ( defined $lldp_entry{lldpRemSysName} ) {
      $conf{topology}{ $if_index }{remote_system} = $lldp_entry{lldpRemSysName};
    }
  }

  foreach my $index ( keys %{ $data{cdp} } ) {
    my %cdp_entry   = %{ $data{cdp}{$index} };

    if ( $cdp_entry{cdpCacheAddressType} ) {
      $conf{topology}{ $index }{address_type} = $cdp_entry{cdpCacheAddressType};
    }
    if ( $cdp_entry{cdpCacheAddress} ) {
      $conf{topology}{ $index }{address}      = $cdp_entry{cdpCacheAddress};
    }
    if ( $cdp_entry{cdpCacheDevicePort} ) {
      $conf{topology}{ $index }{remote_port}  = $cdp_entry{cdpCacheDevicePort};
    }

    if ( defined $cdp_entry{cdpCacheSysName} ) {
      $conf{topology}{ $index }{remote_system} = $cdp_entry{cdpCacheSysName};
    }
    elsif ( defined $cdp_entry{cdpCacheDeviceId} ) {
      $conf{topology}{ $index }{remote_system} = $cdp_entry{cdpCacheDeviceId};
    }
  }

  return 1;
} ## end sub topology

sub vlans {
  ## get vlans from indexed BRIDGE-MIB
  if ( $cisco && keys %{ $data{vlan_member_ports} } ) {
    vlans_cisco();
    return 1;
  }

  ## vlans
  while ( my ($key, $vlan) = each %{ $data{vlans} } ) {
    my %vlan              = %{ $data{vlans}{ $key } };
    my $vlan_name         = $vlan{dot1qVlanStaticName};
    my $portlist_egress   = $vlan{dot1qVlanCurrentEgressPorts};

    unless ( defined $portlist_egress ) {
      next;
    }

    if ( defined $vlan_name ) {
      $conf{vlans}{ $key }{vlan_name} = $vlan_name;
    }

    for my $index ( @{ parse_portlist( $portlist_egress ) } ) {
      my $port_name = $conf{ports}{ $index }{port_name};
      my $vlan_id   = $vlan{dot1qVlanFdbId};

      unless ( defined $port_name && defined $vlan_id ) { next; }
      push @{ $conf{ports}{ $index }{vlan_list} }, $vlan_id;
    }
  }

  return 1;
} ## end sub vlans

sub port_channels {
  foreach my $index ( keys %{ $data{port_channels} } ) {
    my $port_channel;
    my $agg_index  = $data{port_channels}{ $index }{dot3adAggPortSelectedAggID};
    my $pagp_index = $data{port_channels}{ $index }{pagpGroupIfIndex};

    if ( defined $agg_index && $agg_index ne $index && $agg_index ne '0' && defined $data{ports}{ $index } ) {
      $port_channel = $agg_index;
    }
    elsif ( defined $pagp_index && $pagp_index ne $index && $pagp_index ne '0' && defined $data{ports}{ $index } ) {
      $port_channel = $pagp_index;
    }
    else {
      next;
    }

    my %member_port = %{ $data{ports}{ $index } };
    if ( defined $member_port{ifName} ) {
      push @{ $conf{ports}{ $port_channel }{port_channel_members} }, $member_port{ifName};
      push @{ $conf{ports}{ $port_channel }{port_channel_indexes} }, $index;
    }
  }

  return 1;
} ## end sub port_channels

sub switch_physical_info {
  my $switch_model         = "";
  my $switch_os_version    = "";
  my $switch_serial_number = "";
  my @sorted_keys          = sort keys %{ $data{physical_entries} };
  my @classes              = ( 'chassis', 'module' );

 LOOP: foreach my $tested_class ( @classes ) {
    foreach my $phys_entry_index ( @sorted_keys ) {
      my %phys_entry = %{ $data{physical_entries}{ $phys_entry_index } };
      my $class      = defined $phys_entry{entPhysicalClass} ? $phys_entry{entPhysicalClass} : '';

      unless ( $class eq 'chassis' || $class eq 'module' ) { next; }

      my $serial     = defined $phys_entry{entPhysicalSerialNum} ? $phys_entry{entPhysicalSerialNum} : '';
      my $model      = defined $phys_entry{entPhysicalModelName} ? $phys_entry{entPhysicalModelName} : '';
      my $os         = defined $phys_entry{entPhysicalSoftwareRev} ? $phys_entry{entPhysicalSoftwareRev} : '';

      $serial =~ s/\s+$//;
      $model  =~ s/\s+$//;
      $os     =~ s/\s+$//;

      if ( $class eq $tested_class && $serial && $model && $switch_model eq '' && $switch_serial_number eq '' ) {
        $switch_serial_number = $serial;
        $switch_model         = $model;

        if ( $os ) {
          $switch_os_version = $os;
          last LOOP;
        }
      }
      elsif ( $class eq 'module' && $switch_model ne '' && $switch_serial_number ne '' && $os ) {
        $switch_os_version = $os;
        last LOOP;
      }
    }
  }

  return ( $switch_model, $switch_os_version, $switch_serial_number );
} ## end sub switch_physical_info

sub switch_conf {
  my $switch_ip              = $lan_ip;
  my $switch_location        = defined $data{configuration}{sysLocation} ? $data{configuration}{sysLocation} : "";
  my $switch_model           = "";
  my $switch_os_version      = "";
  my $switch_serial_number   = "";
  my $switch_days_up         = defined $data{configuration}{sysUpTime} ? $data{configuration}{sysUpTime} : "";
  my $switch_total_ports     = 0;
  my $switch_unused_ports    = 0;
  my $switch_total_isl_ports = 0;

  ## Days up
  if ( $switch_days_up ) {
    $switch_days_up = (split ':', $switch_days_up)[0];
  }

  ## Port count, unused port count
  while ( my ($index, $port_id) = each %{ $conf{ports} } ) {
    my %port = %{ $conf{ports}{ $index } };
    if ( defined $conf{ports}{ $index }{port_type} && $conf{ports}{ $index }{port_type} eq 'physical' ) {
      $switch_total_ports++;
      $switch_unused_ports++ if ( defined $port{state} && $port{state} eq 'unused' );
    }
  }

  ## Serial number, os version, model
  if ( scalar keys %{ $data{physical_entries} } ) {
    ( $switch_model, $switch_os_version, $switch_serial_number ) = switch_physical_info();
  }

  ## ISL count
  $switch_total_isl_ports = scalar keys %{ $conf{topology} };

  $conf{switch}{switch_name}            = $switch_name;
  $conf{switch}{switch_ip}              = $switch_ip;
  $conf{switch}{switch_location}        = $switch_location;
  $conf{switch}{switch_model}           = $switch_model;
  $conf{switch}{switch_os_version}      = $switch_os_version;
  $conf{switch}{switch_serial_number}   = $switch_serial_number;
  $conf{switch}{switch_days_up}         = $switch_days_up;
  $conf{switch}{switch_total_ports}     = $switch_total_ports;
  $conf{switch}{switch_unused_ports}    = $switch_unused_ports;
  $conf{switch}{switch_total_isl_ports} = $switch_total_isl_ports;

  return 1;
} ## end sub switch_conf

sub save_health_status {
  my $health_status_txt           = "$switch_dir/HEALTH_STATUS.txt";
  my $health_status_txt_time_diff = Xorux_lib::file_time_diff($health_status_txt);
  my @health_status_data;

  # global health status section
  my $main_status    = "OK";
  my $state_suffix  = "ok";
  my $act_timestamp = time();
  my @error_message = ();
  my @states;

  foreach my $index ( sort { $a <=> $b } keys( %{ $data{ports} } ) ) {
    my $port_id          = $conf{ports}{$index}{port_id};
    my $port_type        = $conf{ports}{$index}{port_type};
    my $admin_status     = "";
    my $oper_status      = "";

    unless ( defined $port_id && defined $port_type && $port_type eq 'physical' ) { next; }

    if ( defined $data{ports}{$index}{ifAdminStatus} ) {
      $admin_status = $data{ports}{$index}{ifAdminStatus};
    }
    if ( defined $data{ports}{$index}{ifOperStatus} ) {
      $oper_status = $data{ports}{$index}{ifOperStatus};
    }

    if ( defined $oper_status && $oper_status eq "lowerLayerDown" ) {
      $main_status   = "NOT_OK";
      $state_suffix = "nok";
      push(@error_message, "The operational status of the port $port_id is $oper_status.\n");
    }

    push( @health_status_data, "$port_id,$admin_status,$oper_status\n" );
  }

  if ( !-f $health_status_txt || $health_status_txt_time_diff > 299 ) {
    open( HS, ">$health_status_txt-tmp" ) || Xorux_lib::error( "Couldn't open file $health_status_txt-tmp $!" . __FILE__ . ":" . __LINE__ ) && exit;
    print HS "port,Admin status,Operational status\n";
    print HS "-------------------------------------\n";
    print HS @health_status_data;
    close(HS);
    rename( "$health_status_txt-tmp", "$health_status_txt" );
  } ## end if ( !-f $health_status_txt...)

  if ( $cisco ) {
    if ( modules_status_cisco() eq 'NOT_OK' ) {
      $main_status   = 'NOT_OK';
      $state_suffix  = "nok";
    }
  }

  my $alert_status   = 0;
  my $component_name = $switch_name;
  my $hs_summary_dir = "$tmpdir/health_status_summary";
  $component_name =~ s/\s+//g;

  if ( !-d $hs_summary_dir ) {
    mkdir( $hs_summary_dir, 0755 ) || Xorux_lib::error( "Cannot mkdir $hs_summary_dir: $!" . __FILE__ . ":" . __LINE__ ) && exit;
  }

  if ( !-f "$hs_summary_dir/$component_name.ok" && !-f "$hs_summary_dir/$component_name.nok" ) {
    $alert_status = 1;
  }
  elsif ( -f "$hs_summary_dir/$component_name.ok" ) {
    unlink("$hs_summary_dir/$component_name.ok");
    $alert_status = 1;
  }
  elsif ( -f "$hs_summary_dir/$component_name.nok" ) {
    unlink("$hs_summary_dir/$component_name.nok");
  }

  if ( $state_suffix eq "nok" && $alert_status == 1 ) {
    ### alert health status
    if ( exists &Xorux_lib::health_status_alerting ) {
      Xorux_lib::health_status_alerting( $component_name, \@error_message);
    }
  } ## end if ( $state_suffix eq ...)

  open( MAST, '>', "$hs_summary_dir/$component_name.$state_suffix" ) || Xorux_lib::error( "Couldn't open file $hs_summary_dir/$component_name.$state_suffix $!" . __FILE__ . ":" . __LINE__ ) && exit;
  print MAST "LAN-SWITCH : $switch_name : $main_status : $act_timestamp\n";
  close(MAST);

  return 1;
} ## end sub save_health_status


###########
#  CISCO  #
###########

sub collect_configuration_cisco {
  my $sess = shift;

  if ( $mib_compatibility{"CISCO-PAGP-MIB"} ) {
    SNMP_lib::collect(\%data, $sess, 'port_channels', 'pagpGroupIfIndex');
  }

  ## vlan
  if ( $mib_compatibility{"CISCO-VTP-MIB"} ) {
    SNMP_lib::collect(\%data, $sess, 'vlan_info', 'vtpVlanName', undef, 1);
    SNMP_lib::collect(\%data, $sess, 'vlan_info', 'vtpVlanState', undef, 1);
    SNMP_lib::collect(\%data, $sess, 'vlan_info', 'vtpVlanType', undef, 1);
    SNMP_lib::collect(\%data, $sess, 'vlan_info', 'vtpVlanIfIndex', undef, 1);
  }

  ## routed vlan-iftable relationship
  if ( $mib_compatibility{"CISCO-VLAN-IFTABLE-RELATIONSHIP-MIB"} ) {
    SNMP_lib::collect(\%data, $sess, 'vlan_info', 'cviRoutedVlanIfIndex', undef, 0);
  }

  ## stackport
  if ( $mib_compatibility{"CISCO-STACKWISE-MIB"} ) {
    SNMP_lib::collect(\%data, $sess, 'stackports', 'cswStackPortNeighbor');
    SNMP_lib::collect(\%data, $sess, 'stackports', 'cswStackPortOperStatus');
    SNMP_lib::collect(\%data, $sess, 'stack', 'cswSwitchRole');
    SNMP_lib::collect(\%data, $sess, 'stack', 'cswSwitchState');
  }

  ## FEX
  if ( $mib_compatibility{"CISCO-ETHERNET-FABRIC-EXTENDER-MIB"} ) {
    SNMP_lib::collect(\%data, $sess, 'physical_entries', 'entPhysicalDescr');
    SNMP_lib::collect(\%data, $sess, 'fex_ports', 'cefexBindingExtenderIndex');
    SNMP_lib::collect(\%data, $sess, 'fex_config', 'cefexConfigExtenderName');
  }

  ## VSS
  if ( $mib_compatibility{"CISCO-VIRTUAL-SWITCH-MIB"} ) {
    SNMP_lib::collect(\%data, $sess, 'vss_chassis', 'cvsChassisRole');
    SNMP_lib::collect(\%data, $sess, 'vss_chassis', 'cvsChassisUpTime');
    SNMP_lib::collect(\%data, $sess, 'vss_chassis', 'cvsChassisSwitchID');
    SNMP_lib::collect(\%data, $sess, 'vss_ports', 'cvsVSLConnectOperStatus');
    SNMP_lib::collect(\%data, $sess, 'vss_ports', 'cvsVSLCoreSwitchID');
  }

  ## topology, neighbours
  if ( $mib_compatibility{"CISCO-CDP-MIB"} ) {
    SNMP_lib::collect(\%data, $sess, 'cdp', 'cdpCacheAddressType', undef, 0);
    SNMP_lib::collect(\%data, $sess, 'cdp', 'cdpCacheAddress', \&ip_decimal_format, 0);
    SNMP_lib::collect(\%data, $sess, 'cdp', 'cdpCacheDeviceId', undef, 0);
    SNMP_lib::collect(\%data, $sess, 'cdp', 'cdpCacheDevicePort', undef, 0);
    SNMP_lib::collect(\%data, $sess, 'cdp', 'cdpCacheCapabilities', \&parse_capabilities_cdp, 0);
    SNMP_lib::collect(\%data, $sess, 'cdp', 'cdpCacheSysName', undef, 0);
  }

  ## macs and vlans
  if ( $mib_compatibility{"COMMUNITY-STRING-INDEX"} && keys %{ $data{vlan_info} } ) {
    foreach my $vlan ( keys %{ $data{vlan_info} } ) {
      my %vlan_entry = %{ $data{vlan_info}{ $vlan } };
      unless ( defined $vlan_entry{vtpVlanType} && $vlan_entry{vtpVlanType} eq 'ethernet'
               && defined $vlan_entry{vtpVlanState} && $vlan_entry{vtpVlanState} eq 'operational' ) {
        next;
      }

      my $indexed_sess = SNMP_lib::new_session( { DestHost => $lan_ip }, $vlan );
      SNMP_lib::collect(\%data, $indexed_sess, 'mac_entries', 'dot1dTpFdbAddress', \&mac_format);
      SNMP_lib::collect(\%data, $indexed_sess, 'mac_entries', 'dot1dTpFdbPort');
      SNMP_lib::collect(\%data, $indexed_sess, 'mac_index', 'dot1dBasePortIfIndex');

      my %if_indexes = %{ SNMP_lib::collect(\%data, $indexed_sess, undef, 'dot1dBasePortIfIndex') };
      push @{ $data{vlan_member_ports}{ $vlan } }, values %if_indexes;
    }
  }

  return 1;
} ## end sub collect_configuration_cisco

sub fex_cisco {
  foreach my $index ( keys %{ $data{fex_ports} } ) {
    my $fex_binding = $data{fex_ports}{ $index }{cefexBindingExtenderIndex};

    if ( defined $fex_binding && defined $data{ports}{ $index }{ifName} ) {
      push @{ $conf{fex}{ $fex_binding }{fex_connection} }, $data{ports}{ $index }{ifName};
    }
  }

  foreach my $index ( keys %{ $data{fex_config} } ) {
    my $fex_name = $data{fex_config}{ $index }{cefexConfigExtenderName};

    if ( defined $fex_name ) {
      $conf{fex}{ $index }{fex_name} = $fex_name;
    }

    foreach my $phys_index ( keys %{ $data{physical_entries} } ) {
      my %phys_entry = %{ $data{physical_entries}{ $phys_index } };

      if ( defined $phys_entry{entPhysicalClass} && $phys_entry{entPhysicalClass} eq 'chassis'
           && defined $phys_entry{entPhysicalDescr} && $phys_entry{entPhysicalDescr} =~ /^Fex-$index/
           && $phys_index =~ /$index$/ ) {

        if ( defined $phys_entry{entPhysicalSerialNum} ) {
          $conf{fex}{ $index }{fex_serial} = $phys_entry{entPhysicalSerialNum};
        }

        if ( defined $phys_entry{entPhysicalModelName} ) {
          $conf{fex}{ $index }{fex_model} = $phys_entry{entPhysicalModelName};
        }
      }
    }
  }

  return 1;
} ## end sub fex_cisco

sub vlans_cisco {
  while ( my ($vlan_id, $ports) = each %{ $data{vlan_member_ports} } ) {
    for my $index ( @{ $ports } ) {
      push @{ $conf{ports}{ $index }{vlan_list} }, $vlan_id;
    }
  }

  foreach my $index ( keys %{ $data{vlan_info} } ) {
    my %vlan_entry       = %{ $data{vlan_info}{ $index } };
    my $vlan_name        = $vlan_entry{vtpVlanName};
    my $unrouted_ifindex = $vlan_entry{vtpVlanIfIndex};
    my $routed_ifindex   = $vlan_entry{cviRoutedVlanIfIndex};

    if ( defined $vlan_name ) {
      $conf{vlans}{ $index }{vlan_name} = $vlan_name;
    }
    else {
      next;
    }

    if ( defined $unrouted_ifindex && $unrouted_ifindex ne '0' ) {
      my %interface = %{ $data{ports}{ $unrouted_ifindex } };

      if ( defined $interface{ifName} )    { $conf{vlans}{ $index }{unrouted_interface} = $interface{ifName}; }
      if ( defined $interface{ifAlias} )   { $conf{vlans}{ $index }{unrouted_ifalias}   = $interface{ifAlias}; }
      if ( defined $interface{cfgAlias} )  { $conf{vlans}{ $index }{unrouted_cfgalias}  = $interface{cfgAlias}; }
    }

    if ( defined $routed_ifindex ) {
      my %interface = %{ $data{ports}{ $routed_ifindex } };

      if ( defined $interface{ifName} )    { $conf{vlans}{ $index }{routed_interface} = $interface{ifName}; }
      if ( defined $interface{ifAlias} )   { $conf{vlans}{ $index }{routed_ifalias}   = $interface{ifAlias}; }
      if ( defined $interface{cfgAlias} )  { $conf{vlans}{ $index }{routed_cfgalias}  = $interface{cfgAlias}; }
    }
  }

  return 1;
} ## end sub vlans_cisco

sub stack_cisco {
  my $i = 0;
  my @st_indexes = sort keys %{ $data{stackports} };

  foreach my $index ( sort keys %{ $data{stack} } ) {

    if ( defined $data{stack}{ $index }{cswSwitchRole} ) {
      $conf{stack}{ $index }{role} = $data{stack}{ $index }{cswSwitchRole};
    }

    if ( defined $data{stack}{ $index }{cswSwitchState} ) {
      $conf{stack}{ $index }{state} = $data{stack}{ $index }{cswSwitchState};
    }

    if ( defined $data{physical_entries}{ $index } ) {
      my %phys_entry = %{ $data{physical_entries}{ $index } };

      if ( defined $phys_entry{entPhysicalSerialNum} ) {
        $conf{stack}{ $index }{serial_number} = $phys_entry{entPhysicalSerialNum};
      }

      if ( defined $phys_entry{entPhysicalModelName} ) {
        $conf{stack}{ $index }{model_number} = $phys_entry{entPhysicalModelName};
      }
    }

    for ( my $j = 1; $j < 3; $j++ ) {
      if ( $i < scalar @st_indexes && defined $data{stackports}{ $st_indexes[$i] } ) {
        my %port = %{ $data{stackports}{ $st_indexes[$i] } };
        $i++;

        if ( defined $port{cswStackPortNeighbor} && defined $port{cswStackPortOperStatus}
             && $port{cswStackPortOperStatus} eq 'up' ) {
          $conf{stack}{ $index }{"stackport$j"}         = $port{cswStackPortNeighbor};
          $conf{stack}{ $index }{"stackport$j\_status"} = $port{cswStackPortOperStatus};
        }
      }
    }
  }

  return 1;
} ## end sub stack_cisco

sub vss_cisco {
  foreach my $index ( keys %{ $data{vss_ports} } ) {
    my $switch_id  = $data{vss_ports}{ $index }{cvsVSLCoreSwitchID};
    my $vsl_status = $data{vss_ports}{ $index }{cvsVSLConnectOperStatus};

    unless ( defined $switch_id ) {
      next;
    }

    if ( defined $data{ports}{ $index }{ifName} ) {
      $conf{vss}{ $switch_id }{vss_vsl} = $data{ports}{ $index }{ifName};
    }

    if ( defined $vsl_status ) {
      $conf{vss}{ $switch_id }{vss_vsl_status} = $vsl_status;
    }
  }

  foreach my $index ( keys %{ $data{vss_chassis} } ) {
    my $switch_id      = $data{vss_chassis}{ $index }{cvsChassisSwitchID};
    my $switch_role    = $data{vss_chassis}{ $index }{cvsChassisRole};
    my $switch_uptime  = $data{vss_chassis}{ $index }{cvsChassisUpTime};

    unless ( defined $switch_id ) {
      next;
    }

    if ( defined $switch_uptime ) {
      $conf{vss}{ $switch_id }{vss_uptime} = (split ':', $switch_uptime)[0];
    }

    if ( defined $switch_role ) {
      $conf{vss}{ $switch_id }{vss_role} = $switch_role;
    }

    if ( $data{physical_entries}{ $index } ) {
      my %physical_entry = %{ $data{physical_entries}{ $index } };

      if ( $physical_entry{entPhysicalSerialNum} ) {
        $conf{vss}{ $switch_id }{vss_serial_number} = $physical_entry{entPhysicalSerialNum};
      }

      if ( $physical_entry{entPhysicalName} ) {
        $conf{vss}{ $switch_id }{vss_name} = $physical_entry{entPhysicalName};
      }
    }
  }

  return 1;
} ## end sub vss_cisco

sub modules_status_cisco {
  my $switch_hs_file  = "$switch_dir/HEALTH_STATUS_SWITCH.txt";
  my $main_status     = 'OK';
  my @failed_statuses = ('failed', 'missing', 'mismatchWithParent', 'mismatchConfig','diagFailed', 'dormant',
                        'outOfServiceAdmin','outOfServiceEnvTemp','poweredDown', 'poweredUp', 'powerDenied',
                        'powerCycled', 'okButPowerOverWarning', 'okButPowerOverCritical', 'syncInProgress',
                        'upgrading', 'okButAuthFailed', 'mdr');
  my %switch_status;

  unless ( scalar keys %{ $data{switch_status} } ) {
    if ( -f $switch_hs_file ) {
      unlink($switch_hs_file);
    }

    return 1;
  }

  foreach my $phys_index ( keys %{ $data{switch_status} } ) {
    my $modul_name   = $data{physical_entries}{ $phys_index }{entPhysicalDescr};
    my $modul_serial = $data{physical_entries}{ $phys_index }{entPhysicalSerialNum};
    my $admin_status = $data{switch_status}{ $phys_index }{cefcModuleAdminStatus};
    my $oper_status  = $data{switch_status}{ $phys_index }{cefcModuleOperStatus};

    if ( defined $modul_name && defined $modul_serial && defined $admin_status && defined $oper_status ) {
      $switch_status{modules}{ $phys_index }{modul_name}   = $modul_name;
      $switch_status{modules}{ $phys_index }{modul_serial} = $modul_serial;
      $switch_status{modules}{ $phys_index }{admin_status} = $admin_status;
      $switch_status{modules}{ $phys_index }{oper_status}  = $oper_status;

      foreach my $status ( @failed_statuses ) {
        if ( $oper_status eq $status ) {
          $main_status = 'NOT_OK';
        }
      }
    }
  }

  if ( scalar keys %switch_status ) {
    $switch_status{overall_status} = $main_status;
    Xorux_lib::save_as_json($switch_hs_file, \%switch_status);
  }

  return $main_status;
} ## end sub modules_status_cisco

sub cpu_cisco {
  my $out      = shift;
  my $act_time = time();

  print $out "\n#CPU,act_time,cpu_name,utilization,reserve1,reserve2,switch_name,db_name\n";

  foreach my $index ( sort { $a <=> $b } keys( %{ $data{cpu} } ) ) {
    my %cpu        = %{ $data{cpu}{ $index } };
    my $db_name    = "";
    my $cpu_name   = "";
    my $phys_index = $cpu{cpmCPUTotalPhysicalIndex};

    if ( not defined $phys_index || ( defined $phys_index && ( $phys_index eq '0' || $phys_index eq '1' ) ) ) {
      $cpu_name = 'CPU';
    }
    elsif ( defined $phys_index && defined $data{physical_entries}{ $phys_index }{entPhysicalName} ) {
      $cpu_name = $data{physical_entries}{ $phys_index }{entPhysicalName};
      $cpu_name =~ s/\s+/_/g;
      $cpu_name =~ s/\W//g;
      if ( $cpu_name eq '' ) { next; }
    }
    elsif ( defined $phys_index &&  not defined $data{physical_entries}{ $phys_index }{entPhysicalName} ) {
      $cpu_name = 'CPU';
    }
    else { next; }

    $db_name = $cpu_name . ".lcd";

    my $utilization = '';

    if ( defined $cpu{cpmCPUTotal5minRev} && $cpu{cpmCPUTotal5minRev} =~ /\d{1,3} percent/ ) {
      $utilization = $cpu{cpmCPUTotal5minRev};
      $utilization =~ s/ percent//;
    }

    if ( defined $cpu{rlCpuUtilDuringLast5Minutes} ) {
      $utilization = $cpu{rlCpuUtilDuringLast5Minutes};
    }

    print $out "CPU,$act_time,$cpu_name,$utilization,$switch_name,$db_name\n";
  } ## end foreach my $index ( sort ... cpu )

  return 1;
} ## end sub cpu_cisco

sub memory_cisco {
  my $out      = shift;
  my $act_time = time();

  print $out "\n#MEMORY,act_time,pool_name,utilization,reserve1,reserve2,switch_name,db_name\n";

  foreach my $index ( sort { $a <=> $b } keys( %{ $data{memory} } ) ) {
    my %mem       = %{ $data{memory}{ $index } };
    my $db_name   = "";
    my $pool_name = "";
    my $utilization = '';

    if ( defined $mem{ciscoMemoryPoolName} ) {
      $pool_name = $mem{ciscoMemoryPoolName};
      $pool_name =~ s/\s+/_/g;
      $pool_name =~ s/\W//g;

      if ( defined $mem{ciscoMemoryPoolUsed} && Xorux_lib::isdigit( $mem{ciscoMemoryPoolUsed} )
           && defined $mem{ciscoMemoryPoolFree} && Xorux_lib::isdigit( $mem{ciscoMemoryPoolFree} ) ) {
        $utilization = 100 * ( $mem{ciscoMemoryPoolUsed} / ( $mem{ciscoMemoryPoolUsed} + $mem{ciscoMemoryPoolFree} ) );
      }
      else { next; }
    }
    elsif ( defined $mem{cseSysMemoryUtilization} ) {
      $pool_name   = 'MEM';
      $utilization = $mem{cseSysMemoryUtilization};
      $utilization =~ s/[^0-9]//g;
    }
    else { next; }

    $db_name = $pool_name . ".lmd";

    print $out "MEMORY,$act_time,$pool_name,$utilization,$switch_name,$db_name\n";
  } ## end foreach my $index ( sort ... cpu )

  return 1;
} ## end sub memory_cisco


#############
# Utilities #
#############

sub read_alias_cfg {
  my $alias_file  = "$basedir/etc/alias.cfg";
  my @alias_lines;

  #read alias.cfg
  if ( -f $alias_file ) {
    open( FHA, "< $alias_file" ) || error( "Couldn't open file $alias_file $!" . __FILE__ . ":" . __LINE__ ) && return 0;
    @alias_lines = <FHA>;
    close(FHA);
  }

  foreach my $line ( @alias_lines ) {
    # alias.cfg entries format: LANPORT:cisco-sg200-50:gi1:test
    if ( $line =~ /^LANPORT:([^:]*):([^:]*):([^:]*)$/ && $1 eq $switch_name && $2 ne '' && $3 ne '') {
      foreach my $index ( keys %{ $data{ports} } ) {
        if ( $data{ports}{ $index }{ifName} eq $2 ) {
          $data{ports}{ $index }{cfgAlias} = $3;
        }
      }
    }
  }

  return 1;
} ## end sub read_alias_cfg

sub hex2dec {
  my $num = shift;

  $num =~ s/\s+//g;
  my $hex = "0x$num";

  use Math::BigInt;
  my $dec = Math::BigInt->new($hex);

  return ($dec);
} ## end sub hex2dec

sub bignum {
  my $bnum = shift;

  use Math::BigInt;
  my $num = Math::BigInt->new($bnum);

  return ($num);
} ## end sub bignum

sub parse_portlist {
  my $portlist = shift;
  my $i        = 1;
  my @parts    = split ' ', $portlist;
  my @list_of_ports;

  for my $part ( @parts ) {
    my $dec     = hex( $part );
    my @indexes = $i .. $i + 8;

    push @list_of_ports, map { $dec & ( 128 >> $_ ) ? $indexes[$_] : () } 0 .. 7;
    $i += 8;
  }
  return \@list_of_ports;
} ## end sub parse_portlist

sub ip_decimal_format {
  my $input = shift;
  return $input =~ /([0-9a-fA-F]{2} ?){4}/
    ? join ".", map { hex $_ } split ' ', $input
    : $input;
} ## end sub ip_decimal_format

sub parse_capabilities_cdp {
  my $input        = shift;
  my @capabilities = ( 'Router', 'TB Bridge', 'SR Bridge', 'Switch', 'Host', 'IGMP conditional filtering',
                       'Repeater', 'VoIP Phone', 'Remotely-Managed Device' );
  my @output       = ();

  $input =~ s/ //g;
  unless ( $input =~ /^[0-9a-fA-F]*$/ ) { return \@output; }
  $input = hex( $input );

  for ( my $i = 0; $i < scalar @capabilities ; $i++ ) {
    if ( $input & 1 << $i ) {
      push @output, $capabilities[$i];
    }
  }
  return \@output;
} ## end sub parse_capabilities_cdp

sub parse_capabilities_lldp {
  my $input        = shift;
  my @capabilities = ( 'other', 'repeater', 'bridge', 'wlanAccessPoint', 'router', 'telephone',
                       'docsisCableDevice', 'stationOnly');
  my @output       = ();

  $input =~ s/ //g;
  unless ( $input =~ /^[0-9a-fA-F]*$/ ) { return \@output; }

  my $dec  = hex( $input );
  my $mask = ( 16 ** length $input ) >> 1;
  for ( my $i = 0; $i < scalar @capabilities; $i++ ) {
    if ( $dec & $mask ) {
      push @output, $capabilities[$i];
    }

    $mask >>= 1;
  }

  return \@output;
} ## end sub parse_capabilities lldp

sub mac_format {
  my $input = shift;
  chomp $input;
  $input =~ s/ /:/;

  if ( $input =~ /^(([0-9a-fA-f]{1,2})(:|$)){6}/ ) {
    return join ":", map { length $_ > 1 ? $_ : "0$_" } split /:/, $input;
  }
  else {
    return $input;
  }
} ## end sub mac_format

sub name_format {
  my $name = shift;
  $name =~ s/\//-/g;
  $name =~ s/\s+/_/g;

  return $name;
}

sub port_type {
  my $port_name = shift;
  my $port_type = shift;

  if ( $port_type =~ /^(ethernetCsmacd|fastEther|fastEtherFX|gigabitEthernet)$/  ) { return 'physical'; }
  elsif ( $port_type eq 'tunnel' ) { return 'tunnel'; }
  elsif ( $port_type eq 'propVirtual' || $port_type eq 'ieee8023adLag' ) {
    if ( $port_name =~ /^(VL|Vl|vl)/ ) { return 'vlan'; }
    elsif ( $port_name =~ /^(PO|Po|port-channel)/ ) { return 'port-channel'; }
    elsif ( $port_name =~ /^(Stack|stack)/ ) { return 'stackport'; }
    else { return ''; }
  }
  else { return ''; }
}

sub happened_before {
  my @a = split(/:|\./, shift);
  my @b = split(/:|\./, shift);

  for (my $i = 0; $i < scalar @a; $i++) {
    if ($a[$i] < $b[$i]) {
      return 1;
    }
    if ($a[$i] > $b[$i]) {
      return 0;
    }
  }

  return 0;
} ## end sub happened_before

sub debug {
  my $ref = shift;

  unless ( defined $ref ) {
    $ref = \%data;
  }

  open( DEBUG, ">", "$switch_dir/debug" ) || error( "Couldn't open file $!" . __FILE__ . ":" . __LINE__ ) && exit;
  print DEBUG Dumper( $ref );
  close(DEBUG);
} ## end sub debug
